
{

 *****************************************************************************
 ***                  Easy Computing's "PC UNDERGROUND"                    ***
 ***                  ==================================                   ***
 ***                                                                       ***
 ***                           Unit MOD_SB                                 ***
 ***                                                                       ***
 ***  De unit stelt routines voor het afspelen van MOD-bestanden via de    ***
 ***  Sound Blaster ter beschikking. Omdat de routines zeer tijdkritisch   ***
 ***  zijn, werdebn verschillende mix-routines in 386-assembler geschreven.***
 ***  Daarom werkt de unit niet op een 286-machine. U kunt de MOD-bestanden***
 ***  of door de timer-interrupt laten sturen, of een zogenaamde polling-  ***
 ***  strategie gebruiken. De routines gebruiken voor een 8-stemmig MOD-be-***
 ***  stand op een 486DX-33 bij een weergavefrequentie van 16 kHz minder   ***
 ***  dan 20% van de rekentijd.                                            ***
 ***                                                                       ***
 ***                                                                       ***
 *** Auteur          : Boris Bertelsons  (InspirE)                         ***
 *** Bestandsnaam    : MOD_SB.PAS                                          ***
 *** Laatste versie  : 04.04.1994                                          ***
 *** Versie          : 2.0                                                 ***
 *** Compiler        : Turbo Pascal 6.0 of hoger                           ***
 *****************************************************************************

}

{$F+}
{$G+}
unit mod_sb;

interface uses crt,dos,variab;

 var andycount : word;
     Soundblastergekozeninhoofdprogramma:boolean;

 procedure Soundblastergekozen(waarofniet:boolean);
 procedure voc_pause;
 procedure voc_continue;
 function Init_Sb : boolean;
 {
  Deze functie initialiseert de Sound Blaster. Ze herkent automatisch
  het base-adres en de IRQ, checkt welke Sound Blaster versie geinstal
  leerd is en initialiseert de globale variabelen, die bijvoorbeeld via
  write_sbConfig getoond kunnen worden. Indien de initialisatie geslaagd
  is, wordt de waarde true opgeleverd, zo niet, d waarde false. De luid-
  spreker wordt ingeschakeld. De DMA-ready interrupt wordt naar een
  eigen routine omgelegd.
 }

 procedure dsp_block_sb16(gr,dgr : word;bk : pointer;b1,b2 : boolean);
 {
  Speelt het door bk geadresseerde blok via DMA af
 }

 procedure mod_waitretrace(num : byte);
 {
  Het wachten op een verticale terugslag moet via deze procedure  geschieden
  als ook een MOD wordt afgespeeld (anders schokt het beeld).
 }

 procedure mod_autodetect(what : boolean);
 {
  Zet de speed-herkenning ON/OFF
 }

 procedure mod_SetSpeed(msp : word);
 {
  Initialiseert de variabele "Speed"
 }

 procedure mod_SetLoop(msl : word);
 {
  Initialiseert de variabele "Sound_slepen"
 }

 procedure mod_SetLoopflag(loopen : boolean);
 {
  Zet de flag die aangeeft of een afgelopen MOD-bestand weer van voren
  af aan gespeeld moet worden, ON/OFF
 }

 procedure mod_transpose(transposerwaarde : integer);
 {
  Initialiseert de transposer-waarde. Deze is eventueel voor geluidseffecten
  te gebruiken om een toon omhoog- of omlaag te trekken. NIET wijzigen.
 }

 procedure mod_Samplefreq(Rate : integer);
 {
  Initialiseert de sample-frequentie voor de weergave . Waarde*1000 = frequentie.
  Toegestane waarden zijn 8,10,16,22
 }

 function laad_modbestand(modname : string;ispeed,iloop : integer;freq : byte) : integer;
 {
  Laadt het in modname aangegeven MOD-bestand. Via ispeed en iloop
  kunnen de variabelen "Speed" en "Sound_slepen" ingesteld
  worden, het is echter zinvoller om voor beide waarden de waarde AUTO op te geven.
 }

 procedure einde_mod;
 {
  Stopt het afspelen van een MOD-bestand en verwijdert dit uit geheugen.
 }

 procedure periodiek_on;
 {
  Schakelt het afspelen van een MOD-bestand aan. Moet voor het starten aangeroepen
  worden.
 }

 procedure periodiek_off;
 {
  Onderbreekt het afspelen van een MOD-bestand.Het MOD-bestand blijft in het
  geheugen en kan via periodiek_on weer gestart worden.
 }

 procedure write_sbConfig;
 {
  Toont de gevonden configuratie op het scherm (tekstmode!). Alternatief voor
  direct access van de betreffende variabelen.
 }

 procedure calculate_music;
 {
  Berekend een stuk muziek. Wordt periodiek of in een loop aangeroepen.
 }

 procedure Filter_aan;
 {
  Schakelt het bass-filter aan.
 }

 procedure Filter_MID;
 {
  Schakelt het filter naar normaal bedrijf.
 }
 procedure Filter_uit;
 {
  Schakelt het bass-filter uit.
 }

 procedure Set_Balance(Waarde : byte);
 {
  De procedure stelt de balans volgens de overgedragen waarde in.
  De waarde 0 betekent geheel links, de waarde 24 geheel links. De waarde 12
  betekent midden.
 }

 procedure Set_Volume(Waarde : byte);
{
  Deze procedure stelt de luidsterkte voor de gehele weergave in
  (Master Volume). De toegestane waarden liggen tussen 0 en 31.
 }

 procedure wr_dsp_sb16(v : byte);
{
  Schrijft de overgedragen variabelen in het Soundblaster-Register
 }

 function init_The_Mod : boolean;
 {
  Initialiseert de MOD-routines. Reset de SB en initialiseert de
  bevodigde variabelen.
 }

 FUNCTION Reset_sb16 : BOOLEAN;
{
  Reset de SB-kaart. Levert TRUE op, indien gelukt.
 }

 procedure Set_Timeconst_sb16(tc : byte);
{
  Zet de timer-constante die volgens de formule
  tc := 256-(1.000.000 / frequentie) berekend wordt.
 }

 procedure Fade_Musix_out;
 {
  Laadt de luidsterkte langzaam teruglopen. Bijvoorbeeld voor gebruik aan het
  einde van een programma.
 }

 procedure Afspelen_Sb(Segm,Offs,dgr,dsize : word);
{
  Speelt het aangegeven blok via het DMA af. Tbv SB / SB Pro
 }

 procedure Afspelen_Sb16(Segm,Offs,dgr,dsize : word);
{
  Speelt het aangegeven blok via het DMA af. Tbv SB 16
 }

 procedure init_data;
 {
  Initialiseert de variablen van de unit
 }

 Procedure Init_Voc(filename : string);
{
  De procedure start de weergave van de in de filenaam overgedragen
  VOC-file. Parallelle weergave van een VOC en een MOD is niet mogelijk.
 }

 procedure voc_done;
 {
  Stopt het afspelen van een VOC-bestand.
 }

const bpm : byte = 125;

var SaveExitProc : Pointer;             { nodig wegens eigen exitproc}
    mycli : byte;                       { flag, of Soundberekening   }
                                        { actief                     }
    music_played : boolean;             { flag, TRUE indien muziek   }
                                        { gespeeld wordt             }
implementation

const
 Speed3 : word = 58;
 Loop3  : word = 42;

Var
 Toonhoogte_stem : array[1..8] of word;
 toonhoogte : word;                     { de Waarde van de toonhoogte}
                                        { zoals deze in MOD-file staat}
 doel : pt;                             { afspeelbuffer in pt-formaat }
 Modp  : pointer;                       { pointer naar Rm_Song        }
 noot : array[1..8] of byte;
 altx,alty : integer;
 actuele_stem : word;

 Mixed_data   : pointer;
 Mixed_data_st : pointer;

{$L modmix}
procedure lege_stem; external;
{procedure stem_normaal(actuele_stem : word); external;}
procedure stem_normaal; external;
procedure stem_normaal_st; external;

var Portamento_Up_stem : array[1..8] of longint;
    Portamento_Do_stem : array[1..8] of longint;

var Mixingprocs         :  array[1..8] of pointer;
    Stemleeg          : pointer;

    effect_stem     : array[1..8] of byte;

    Lengte_stem       : array[1..8] of longint;
    Loop_Lengte_stem  : array[1..8] of longint;
    Positie_stem     : array[1..8] of longint;
    Loop_Start_stem   : array[1..8] of longint;
    Segment_stem      : array[1..8] of longint;
    Nootvol_stem       : array[1..8] of longint;
    Incval_stem       : array[1..8] of longint;

var nothin_done_count : word;

    var bsw : boolean;
      vocb1,
      vocb2,
      buffer1,
      buffer2 : pointer;
    shiftfactor,
    shiftfactor_stereo : word;

{

 **************************************************************************
 ***                                                                    ***
 ***                Routines  voor de  timer - handling                 ***
 ***                                                                    ***
 **************************************************************************

}

procedure TimerAan(Proc : pointer; Freq : word);
var iteller : word;
    oldv : pointer;
begin;
 asm cli end;
 iteller := 1193180 DIV Freq;
 Port[$43] := $36;
 Port[$40] := Lo(iteller);
 Port[$40] := Hi(iteller);

 Getintvec(8,OldV);
 setintvec(OldTimerInt,OldV);
 SetIntVec(8,Proc);
 old_tteller := 1;
 secteller  := 0;
 oldintteller := 0;
 asm sti end;
end;


procedure Timeruit;
var oldv : pointer;
begin;
  asm cli end;
  Port[$43] := $36;
  Port[$40] := 0;
  Port[$40] := 0;
  GetIntVec(OldTimerInt,OldV);
  SetIntVec(8,OldV);
  asm sti end;
end;

procedure NieuweTimer; interrupt;
var dr : registers;
begin;
 inc(nothin_done_count);
 inc(Secteller);
 inc(Oldintteller);
 if Oldintteller = 58 then begin;
   Oldintteller := 0;
   intr(Oldtimerint,dr);
 end;
 if Secteller= timer_per_second then begin;
   Secteller := 0;
   inc(andycount);
   inc(Speelsec);
   if Speelsec = 60 then begin;
     inc(Speelmin);
     Speelsec := 0;
   end;
 end;
 if not in_retrace then calculate_Music;
 Port[$20] := $20;
end;


{

 **************************************************************************
 ***                                                                    ***
 ***              Routines voor de aansturing van de Soundblaster       ***
 ***                                                                    ***
 **************************************************************************

}


procedure wr_dsp_sb16(v : byte);
begin;
  while port[dsp_adr+$c] >= 128 do ;
  port[dsp_adr+$c] := v;
end;

FUNCTION SbReadByte : BYTE;
begin;
  while port[dsp_adr+$a] = $AA do ;     { wachten tot DSP ready        }
  SbReadByte := port[dsp_adr+$a];       { waarde schrijven             }
end;

procedure SBreset;
VAR bt,ct, stat : BYTE;
begin;
  PORT[dsp_adr+$6] := 1;                { dsp_adr+$6 = Resetfunctie}
  FOR ct := 1 TO 100 DO;
  PORT[dsp_adr+$6] := 0;
  bt := 0;
  repeat
    ct := 0;
    repeat
      stat := port[dsp_adr + $E];
    until (ct > 8000) or (stat >= 128);
    inc(bt);
  until (bt > 100) or (port[dsp_adr + $A] = $AA);
end;

FUNCTION Reset_sb16 : BOOLEAN;
CONST  ready = $AA;
VAR ct, stat : BYTE;
BEGIN
  PORT[dsp_adr+$6] := 1;                { dsp_adr+$6 = Resetfunctie}
  FOR ct := 1 TO 100 DO;
  PORT[dsp_adr+$6] := 0;
  stat := 0;
  ct   := 0;                            { vergelijk ct < 100, omdat  }
  WHILE (stat <> ready)                 { initialisatie ca.          }
  AND   (ct < 100)      DO BEGIN        { 100ms duurt               }
    stat := PORT[dsp_adr+$E];
    stat := PORT[dsp_adr+$a];
    INC(ct);
  END;
  Reset_sb16 := (stat = ready);
END;

FUNCTION Detect_Reg_sb16 : BOOLEAN;
VAR
  Port, Lst : WORD;
BEGIN
 Detect_Reg_sb16 := SbRegDetected;
 IF SbRegDetected THEN EXIT;            { exit indien geinitialiseerd }
 Port := Startport;                     { mogelijke SB-adressen tussen}
 Lst  := Endport;                       { $210 en $280 !              }
 WHILE (NOT SbRegDetected)
 AND   (Port <= Lst)  DO BEGIN
   dsp_adr := Port;
   SbRegDetected := Reset_sb16;
   IF NOT SbRegDetected THEN
     INC(Port, $10);
 END;
 Detect_Reg_sb16 := SbRegDetected;
END;

PROCEDURE Write_Mixer(Reg, Val: BYTE);
begin;
 Port[dsp_adr+$4] := Reg;
 Port[dsp_adr+$5] := Val;
END;


FUNCTION Read_Mixer(Reg: BYTE) : BYTE;
begin;
  Port[dsp_adr+$4] := Reg;
  Read_Mixer := Port[dsp_adr+$5];
end;

procedure Filter_aan;
var aux : byte;
begin;
 if sb16detected then begin;
   write_Mixer(68,64);                  { treble omlaag              }
   write_Mixer(69,64);
   write_Mixer(70,255);                 { Bass full Power !          }
   write_Mixer(71,255);                 { Bass full Power !          }
 end else begin;
   aux := read_Mixer($0c);              { laagfilter                 }
   aux := aux or 8;
   Write_Mixer($0c,aux);
   aux := read_Mixer($0e);              { filter aan                 }
   aux := aux AND 2;
   write_Mixer($0e,aux);
 end;
end;

procedure Filter_MID;
var aux : byte;
begin;
 if sb16detected then begin;
   write_Mixer(68,160);                 { treble omlaag              }
   write_Mixer(69,160);
   write_Mixer(70,192);                 { Bass full Power !          }
   write_Mixer(71,192);                 { Bass full Power !          }
 end else begin;
   aux := read_Mixer($0e);              { filter uit                 }
   aux := aux OR 32;
   write_Mixer($0e,aux);
 end;
end;

procedure Filter_uit;
var aux : byte;
begin;
 if sb16detected then begin;
   write_Mixer(68,192);                 { standaardinstelling       }
   write_Mixer(69,192);
   write_Mixer(70,160);
   write_Mixer(71,160);
 end else begin;
   aux := read_Mixer($0c);              { hoogfilter                }
   aux := aux OR 247;
   Write_Mixer($0c,aux);
   aux := read_Mixer($0e);              { filter aan                }
   aux := aux AND 2;
   write_Mixer($0e,aux);
 end;
end;

procedure Set_Balance(Waarde : byte);
Var left,right : byte;
begin;
 if Sb16Detected then begin;
   left  := 12;
   right := 12;
   if Waarde < 12 then right := waarde;
   if Waarde > 12 then left  := 24-Waarde;
   write_Mixer(50,(left  shl 4));
   write_Mixer(51,(right shl 4));
 end else begin;
  Waarde := Waarde SHR 1;
  case Waarde of
     0..6 : begin;
               write_Mixer(02,(7 shl 5)+(Waarde shl 1));
             end;
       07 : begin;
               write_Mixer(02,(7 shl 5)+(7 shl 1));
             end;
       08..13 : begin;
               write_Mixer(02,((13-Waarde) shl 5)+(7 shl 1));
                end;
       end;
 end;
end;

procedure Set_Volume(Waarde : byte);
begin;
  if sb16detected then begin;
    write_Mixer(48,(Waarde shl 3));
    write_Mixer(49,(Waarde shl 3));
  end else begin;
    if MixerDetected then begin;
      Waarde := Waarde Shr 2;
      write_Mixer($22,(waarde shl 5) + (waarde shl 1));
    end else begin;
      outvolume := Waarde shl 1;
    end;
  end;
end;

procedure reset_Mixer; assembler;
asm
  mov dx,dsp_adr+$4
  mov al,0
  out dx,al
  mov cx,50
@loop:
  loop @loop
  inc dx
  out dx,al
end;

FUNCTION Detect_Mixer_sb16 : BOOLEAN;
VAR SaveReg : WORD;
    NewReg  : WORD;
BEGIN
  Detect_Mixer_sb16 := MixerDetected;
  IF (NOT SbRegDetected)                { Stoppen als geen Sound-    }
  OR MixerDetected THEN EXIT;           { blaster-kaart aanwezig is  }
                                        { of de mixer-chip reeds     }
                                        { geinitialiseerd is         }
  Reset_Mixer;
  SaveReg := Read_Mixer($22);           { registers bewaren          }
  Write_Mixer($22, 243);                { als de geschreven waarde   }
  NewReg  := Read_Mixer($22);           { met de teruggelezen waarde }
                                        { overeenkomt, dan is        }
                                        { access mogelijk en dus     }
                                        { een mixer aanwezig         }
  IF NewReg = 243 THEN begin;
    MixerDetected := TRUE;
    STEREO := True;
  end;
  Write_Mixer($22, SaveReg);            { oude register terug        }
  Detect_Mixer_sb16 := MixerDetected;
END;

PROCEDURE SbGetDSPVersion;
VAR i : WORD;
    t : WORD;
    s : STRING[2];
BEGIN
  wr_dsp_sb16($E1);                     { $E1 = opvragen versie      }
  SbVersMaj := SbReadByte;
  SbVersMin := SbReadByte;
  str(SbVersMaj, SbVersStr);
  SbVersStr := SbVersStr + '.';
  str(SbVersMin, s);
  if SbVersMin > 9 then
    SbVersStr := SbVersStr +       s
  else
    SbVersStr := SbVersStr + '0' + s;
END;

function wrt_dsp_adr_sb16 : string;
begin;
  case dsp_adr of
    $210 : wrt_dsp_adr_sb16 := '210';
    $220 : wrt_dsp_adr_sb16 := '220';
    $230 : wrt_dsp_adr_sb16 := '230';
    $240 : wrt_dsp_adr_sb16 := '240';
    $250 : wrt_dsp_adr_sb16 := '250';
    $260 : wrt_dsp_adr_sb16 := '260';
    $270 : wrt_dsp_adr_sb16 := '270';
    $270 : wrt_dsp_adr_sb16 := '280';
   END;
end;

function wrt_dsp_irq : string;
begin;
  case dsp_irq of
     $2 : wrt_dsp_irq := '2 h';
     $3 : wrt_dsp_irq := '3 h';
     $5 : wrt_dsp_irq := '5 h';
     $7 : wrt_dsp_irq := '7 h';
    $10 : wrt_dsp_irq := '10 h';
   END;
end;

procedure Set_Timeconst_sb16(tc : byte);
begin;
  wr_dsp_sb16($40);                     { $40 = zet sample rate    }
  wr_dsp_sb16(tc);
end;

procedure Soundblastergekozen(waarofniet:boolean);
begin
Soundblastergekozeninhoofdprogramma:=waarofniet;
end;

procedure test_overdracht;
begin;
   fillchar(buffer1^,3000,127);
   blockgrootte := 2000;
   laatste_weergave := true;
   Sampling_Rate := 211;
   dsp_block_sb16(blockgrootte,blockgrootte,buffer1,true,false);
   delay(100);
end;

procedure write_sbConfig;
begin;
  clrscr;
  if SbRegDetected then begin;
    writeln('Soundkaart op base ',wrt_dsp_adr_sb16,'h met IRQ ',
    wrt_dsp_irq,' gevonden.');
  end else begin;
    writeln('Geen Soundblaster-compatibele kaart gevonden !');
  end;
  if MixerDetected then begin;
    writeln('Mixer - Chip gevonden');
    if SbVersMaj < 4 then
      writeln('De gevonden kaart is',
           ' een Soundblaster Pro of compatibel')
    else
      writeln('De gevonden kaart is',
      ' een Soundblaster 16 ASP of compatibel');
  end else begin;
    writeln('De gevonden kaart is',
    ' een Soundblaster of compatibel');
  end;
  writeln('Het versienummer luidt ',SbVersStr);
end;

procedure Exit_Sb16;
begin;
  setintvec($8+dsp_irq,oldint);         { oude interrupt herstellen    }
  port[$21] := Port[$21] or irqmsk;     { en maskering naar oude waarde}
  port[dsp_adr+$c] := $d3;              { terug                        }
  Port[$20] := $20;
  wr_dsp_sb16($D0);
end;

procedure Afspelen_Sb16(Segm,Offs,dgr,dsize : word);
var li : word;
begin;
  port[$0A] := dma_ch+4;                { DMA-kanaal locken          }
  Port[$0c] := 0;                       { adres van de buffer        }
  Port[$0B] := $48+dma_ch;              { voor de geluidsweergave    }
  Port[dma_adr[dma_ch]] := Lo(offs);    { op de DMA-controller       }
  Port[dma_adr[dma_ch]] := Hi(offs);
  Port[dma_wc[dma_ch]] := Lo(dgr-1);    { omvang van het blok       }
  Port[dma_wc[dma_ch]] := Hi(dgr-1);    { op de DMA-controller      }
  Port[dma_page[dma_ch]] := Segm;
  if sb16_outputlengte <> dsize then begin;
    wr_dsp_sb16($C6);                   { DSP-opdracht 8-Bit via DMA }
    if stereo then begin;               { voor SB16 alleen om te starten ! }
      wr_dsp_sb16($20);
{      write('');}
    end else
      wr_dsp_sb16($00);
    wr_dsp_sb16(Lo(dsize-1));           { omvang blok op             }
    wr_dsp_sb16(Hi(dsize-1));           { de DSP                     }
    sb16_outputlengte := dsize;
  end else begin;
    wr_dsp_sb16($45);                   { DMA continue SB16 8-Bit    }
  end;
  Port[$0A] := dma_ch;                  { DMA-kanaal vrijgeven        }
end;

procedure Afspelen_Sb(Segm,Offs,dgr,dsize : word);
var li : word;
begin;
  port[$0A] := dma_ch+4;
  Port[$0c] := 0;
  Port[$0B] := $48+dma_ch;
  Port[dma_adr[dma_ch]] := Lo(offs);
  Port[dma_adr[dma_ch]] := Hi(offs);
  Port[dma_wc[dma_ch]] := Lo(dgr-1);
  Port[dma_wc[dma_ch]] := Hi(dgr-1);
  Port[dma_page[dma_ch]] := Segm;
  wr_dsp_sb16($14);
  wr_dsp_sb16(Lo(dsize-1));
  wr_dsp_sb16(Hi(dsize-1));
  Port[$0A] := dma_ch;
end;

procedure Afspelen_SbPro(Segm,Offs,dgr,dsize : word);
var li : word;
begin;
  port[$0A] := dma_ch+4;
  Port[$0c] := 0;
  Port[$0B] := $48+dma_ch;
  Port[dma_adr[dma_ch]] := Lo(offs);
  Port[dma_adr[dma_ch]] := Hi(offs);
  Port[dma_wc[dma_ch]] := Lo(dgr-1);
  Port[dma_wc[dma_ch]] := Hi(dgr-1);
  Port[dma_page[dma_ch]] := Segm;

  wr_dsp_sb16($48);
  wr_dsp_sb16(Lo(dsize-1));
  wr_dsp_sb16(Hi(dsize-1));
  wr_dsp_sb16($91);
  Port[$0A] := dma_ch;
end;

procedure dsp_block_sb16(gr,dgr : word;bk : pointer;b1,b2 : boolean);
var l : longint;
    pn,offs : word;
    hbyte : byte;
    a : word;
    OldV,NewV,Aux : byte;
    stereoreg : byte;
    sr : word;
    samps : byte;
begin;
  PLAYING_MOD := b1;
  PLAYING_VOC := b2;

  dsp_rdy_sb16 := false;
  l := 16*longint(pt(bk).sgm)+pt(bk).ofs;
  pn := pt(l).sgm;
  offs := pt(l).ofs;

  if PLAYING_MOD then begin;
    set_timeconst_sb16(Sampling_Rate);
    if sb16Detected then begin;
      if stereo then  begin;
        Afspelen_Sb16(pn,offs,dgr*2,gr*2);
{        write('S');}
      end else
        Afspelen_Sb16(pn,offs,dgr,gr);
    end else begin;
      if stereo then begin;
        SR := word(-1000000 DIV (Sampling_Rate-256));
        SR := SR * 2;
        Samps := 256 - (1000000 DIV SR);
        set_timeconst_sb16(Samps);
        Afspelen_SbPro(pn,offs,dgr*2,gr*2);
      end else
        Afspelen_Sb(pn,offs,dgr,gr);
    end;
  end;

  if PLAYING_VOC then begin;
    sb16_outputlengte := 0;
    set_timeconst_sb16(vblock.SR);
    if sb16Detected then begin;
      if stereo then begin;
        Afspelen_Sb16(pn,offs,dgr,gr);
      end else begin;
        Afspelen_Sb16(pn,offs,dgr,gr);
      end;
    end else begin;
      if stereo then begin;
        Afspelen_SbPro(pn,offs,dgr,gr);
      end else begin;
        Afspelen_Sb(pn,offs,dgr,gr);
      end;
    end;
  end;
end;


{

 **************************************************************************
 ***                                                                    ***
 ***                Routines voor het afspelen van MOD-bestanden        ***
 ***                                                                    ***
 **************************************************************************

}

procedure get_pctunel(hoogte : word;Var vk : longint);
{
 De procedure bepaalt uit de overgedragen toonhoogte (zoals deze in
 het MOD-bestand staat) de voor de frequentiemanipulatie benodigde
 plaatsen voor- en achter de komma.
}
var nct : byte;
    gevonden : boolean;
begin;
 nct := 1;
 gevonden := false;
 while (nct <= 70) and not gevonden do  {tot gevonden of laatste}
 begin;                                 {waarde in tabel}
   if hoogte > Modoctaaf[nct] then
     gevonden := true;
   inc(nct);
 end;
 if gevonden then begin;
   vk  := Incfacts[nct-tpw+12];
 end else begin;
   vk  := 0;                            {waarde uit tabel halen}
 end;
end;


{function T_hoogte(Nr : word) : integer;
begin;
 T_hoogte := Modoctaaf[nr];
end;}

function Noot_nr(hoogte : word) : integer;
var nct : byte;
    gevonden : boolean;
begin;
 nct := 1;
 gevonden := false;
 while (nct <= 70) and not gevonden do  {tot gevonden of laatste}
 begin;                                 {waarde in tabel}
   if hoogte > Modoctaaf[nct] then
     gevonden := true;
   inc(nct);
 end;
 if gevonden then begin;
   Noot_nr := nct-1;
 end else begin;
   Noot_nr := -1;
 end;
end;

function Nootvolume(Stm : byte) : byte;
begin;
 Nootvolume := Rm_Song[mli,Stm,4];
end;

var mixed_posi : word;
    calc_size : word;

procedure binnen_loop_4;
{
 Hier begint de mixing van de muziekgegevens. De buffer
 wordt daarbij met de berekende data gevuld. Dit is de
 monoversie van de routine.
}
begin;
  calc_size := blockgrootte;
  if bsw then
    doel := pt(buffer1)
  else
    doel := pt(buffer2);
  fillchar(mixed_data^,8000,128);
asm

    mov cx,1
@stemmen_loop:
    mov mixed_posi,0
    mov actuele_stem,cx
    mov si,cx
    dec si
    shl si,2
    call dword ptr Mixingprocs[si]

    inc cx
    cmp cx,stemmen
    jbe @stemmen_loop

    mov mixed_posi,0
    mov cx,calc_size
@Mixed_2_blk:
    les di,mixed_data
    add di,mixed_posi
    mov ax,es:[di]
    push cx
    mov cx,shiftfactor
    shr ax,cl
    pop cx
    add mixed_posi,2

    mov bx,doel.sgm             {byte naar doel schrijven}
    mov es,bx
    mov bx,doel.ofs
    mul outvolume
    shr ax,6
    mov es:[bx],al
    inc doel.ofs

    loop @mixed_2_blk
end;
end;

procedure binnen_loop_4_stereo;
{
 Hier begint de mixing van de muziekgegevens. De buffer
 wordt daarbij met de berekende data gevuld. Dit is de
 stereoversie van de routine.
}
begin;
  calc_size := blockgrootte;
  if bsw then
    doel := pt(buffer1)
  else
    doel := pt(buffer2);
  fillchar(mixed_data^,8000,128);
  fillchar(mixed_data_st^,8000,128);
asm
    mov cx,1
@stemmen_loop:
    mov mixed_posi,0
    mov actuele_stem,cx
    mov si,cx
    dec si
    shl si,2
    call dword ptr Mixingprocs[si]

    inc cx
    cmp cx,stemmen
    jbe @stemmen_loop

    mov mixed_posi,0
    mov cx,calc_size
@Mixed_2_blk:
    les di,mixed_data
    add di,mixed_posi
    mov ax,es:[di]
    push cx
    mov cx,shiftfactor_stereo
    shr ax,cl
    pop cx

    mov bx,doel.sgm             {byte naar doel schrijven}
    mov es,bx
    mov bx,doel.ofs
    mul outvolume
    shr ax,6
    mov es:[bx],al
    inc doel.ofs

    les di,mixed_data_st
    add di,mixed_posi
    mov ax,es:[di]
    push cx
    mov cx,shiftfactor_stereo
    shr ax,cl
    pop cx
    add mixed_posi,2

    mov bx,doel.sgm             {byte naar doel schrijven}
    mov es,bx
    mov bx,doel.ofs
    mul outvolume
    shr ax,6
    mov es:[bx],al
    inc doel.ofs

    loop @mixed_2_blk
end;
end;

procedure start_mixen_4;
var rdiff : real;
    dummy : byte;
    var li : integer;
begin;
 for li := 1 to Stemmen do begin;
   if noot[li] <> 0 then begin;
     Toonhoogte_Stem[li] := (Rm_Song[mli,li,1] and $0F)*256+Rm_Song[mli,li,2];
     get_pctunel(toonhoogte_Stem[li],Incval_Stem[li]);
   end;

   ls[li] := loop_s[In_st[li]];
   ll[li] := loop_l[In_st[li]];
   if ll[li] > 30 then inl[li] := ll[li]+ls[li];
   Loop_Lengte_Stem[li] := ll[li];
   Loop_Start_Stem[li]  := ls[li];

   case effect_Stem[li] of
     1 : begin;
            inc(Incval_Stem[li],Portamento_up_Stem[li]);
         end;
     2 : begin;
            inc(Incval_Stem[li],Portamento_do_Stem[li]);
         end;
   end;
 end;
end;


procedure effect_handling(li : integer);
var idx : word;
    Portamento_Speed : word;
    Startnoot,
    eindnoot : word;
    startinc,
    eindinc : longint;

begin;
 if Rm_Song[mli,li,3] and $0F <= 15 then begin;
   Eff[li] := 0;
   case (Rm_Song[mli,li,3] and $0F) of
      01 : begin;
             effect_Stem[li] := 1;
             Portamento_Speed := Rm_Song[mli,li,4];
             Startnoot := noot_nr(toonhoogte_stem[li]);
             eindnoot := Startnoot+Portamento_Speed;
             get_pctunel(Modoctaaf[Startnoot],Startinc);
             get_pctunel(Modoctaaf[Eindnoot],eindinc);
             Portamento_up_Stem[li] := round((eindinc - Startinc) / playspeed);
           end;
      02 : begin;
             effect_Stem[li] := 2;
             Portamento_Speed := Rm_Song[mli,li,4];
             Startnoot := Noot_nr(toonhoogte_stem[li]);
             eindnoot   := Startnoot-Portamento_Speed;
             get_pctunel(Modoctaaf[Startnoot],Startinc);
             get_pctunel(Modoctaaf[eindnoot],eindinc);
             Portamento_do_Stem[li] := round((eindinc - Startinc) / playspeed);
           end;
       9 : begin;  { Sample offset }
             Positie_Stem[li] := longint(Rm_Song[mli,li,4]) shl 24;
           end;
      13 : begin;
             mli := 64;
     end;
      11 : begin;
             mli := 64;
             mlj := Rm_Song[mli,li,4];
     end;
      12 : begin;
             Nootvol_Stem[li]  := Nootvolume(li);
           end;
      14 : begin;
             case (Rm_Song[mli,li,4] shr 4) of
                12 : begin;
                       inl[li]    := 0;
                       Nootvol_Stem[li]:= 0;
                       inp[li]    := 0;
                       Pnk[li]    := 0;
                     end;
             end;
           end;
      15 : begin;
             idx := Rm_Song[mli,li,4];
             if idx <= $f then begin;
               Playspeed   := idx;
               Speed       := Playspeed*105 div 10;
               blockgrootte := Speed * Sound_slepen;
             end else begin;
               bpm := idx;
               mod_SetLoop(Sampling_Frequentie div (BPM * 4));
               Speed := Playspeed*105 div 10;
               blockgrootte := Speed * Sound_slepen;
             end;
             if blockgrootte < 40 then blockgrootte := 40;
             if blockgrootte > 4000 then blockgrootte := 4000;
           end;
    end;
  end;
end;

procedure nmw_all_4;
const stereoprocs : array[1..8] of pointer =
  (@Stem_normaal,@Stem_normaal,@Stem_normaal_st,@Stem_normaal_st,
   @Stem_normaal,@Stem_normaal,@Stem_normaal_st,@Stem_normaal_st);
var idx : byte;
    li : integer;
begin;
   inc(mli);
   if mli > 64 then mli := 1;
   if mli = 1 then begin;
     inc(mlj);
     if mlj > Songlengte then begin;
       if mloop then begin;
         mlj := 1;
         move(rm[song[mlj]] ^,Rm_Song,2048);
       end else begin;
         asm
           call [periodiek_onderbreken]
         end;
         music_uit := true;
         Mod_afgelopen := true;
       end;
     end else begin;
         move(rm[song[mlj]] ^,Rm_Song,2048);
     end;
   end;

  for li := 1 to Stemmen do begin;
    effect_Stem[li] := 0;
    noot[li] := (Rm_Song[mli,li,1] AND $F0)+((Rm_Song[mli,li,3] AND $F0) shr 4);
    if noot[li] <> 0 then begin;
      if stereo then begin;
        Mixingprocs[li] := stereoprocs[li];
      end else begin;
        Mixingprocs[li] := @Stem_normaal;
      end;
      Noot_uitslag[li] := 500;
      In_St[li]  := noot[li];
      inst[li]  := Ptr(pt(Samp[In_St[li]]).sgm,pt(Samp[In_St[li]]).ofs);
      Lengte_Stem[li]      := sam_l[In_St[li]];
      Positie_Stem[li]    := 0;
      Nootvol_Stem[li]      := inst_vol[in_St[li]];
      Segment_Stem[li]     := seg(inst[li]^);
    end;
    effect_handling(li);
  end;

end;


procedure initialiseer_mengen;
begin;
  asm
    call [meng_proc]
  end;
end;

FUNCTION ConvertString(Source : Pointer; Size : BYTE):String;
VAR
   WorkStr : String;
BEGIN
   Move(Source^,WorkStr[1],Size);
   WorkStr[0] := CHR(Size);
   ConvertString := WorkStr;
END;

function init_Song : boolean;
const id1 : string = 'FLT4';
      id2 : string = 'M.K.';
      id3 : string = '8CHN';
var rmod : file;
    sgr : word;                         {sample-grootte}
    inststart : longint; {plaats in bestand, waar sample-data beginnen}
    datgr : longint;                    {omvang MOD gegevens}
    Mkg : array[1..4] of char;          {indentificatie mod type}
    hulpptr : ^byte;
    strptr : pointer;
    idch : array[1..4] of char;
    idstr : string;
    instance : byte;
    idx : integer;
begin;
 In_St[1] := 0;
 In_St[2] := 0;
 In_St[3] := 0;
 In_St[4] := 0;
 In_St[5] := 0;
 In_St[6] := 0;
 In_St[7] := 0;
 In_St[8] := 0;
 for mlj := 0 to 128 do
   song[mlj] := 0;
 {$I-}
 assign(rmod,Mod_Name);
 reset(rmod,1);
 {$I+}
 if IOresult <> 0 then begin;
   init_song := false;
   exit;
 end;
 if moddatgrootte <> 0 then datgr := moddatgrootte else
 datgr := filesize(rmod);
 inststart := datgr;
 seek(rmod,1080);
 blockread(rmod,idch,4);
 idstr := idch;
 if (idstr <> id1) and (idstr <> id2)
 and (idstr <> id3) then begin;
   instance := 15;
 end else begin;
   instance := 31;
 end;

 if instance = 31 then begin;   {31 stemmen herkend}
 for mlj := 1 to 31 do begin;
  idx := mlj;
  seek(rmod,msp+42+(idx-1)*30);
  blockread(rmod,sgr,2);
  sgr := swap(sgr) * 2;
  if sgr <> 0 then inststart := inststart - sgr;
  Sam_l[idx] := sgr;
  seek(rmod,msp+45+(idx-1)*30);
  blockread(rmod,inst_vol[idx],1);
  blockread(rmod,loop_s[idx],2);
  blockread(rmod,loop_l[idx],2);
  loop_s[idx] := swap(loop_s[idx])*2;
  loop_l[idx] := swap(loop_l[idx])*2;
 end;

 seek(rmod,msp+1080);
 blockread(rmod,Mkg,4);
 if pos('8CHN',Mkg) <> 0 then begin;
   Pattgrootte         := 2048;
   Stemmen            := 8;
   shiftfactor        := 3;
   shiftfactor_stereo := 3;
 end else begin;
 {4-stemmige MOD-gegevens }
   Pattgrootte    := 1024;
   Stemmen        := 4;
   shiftfactor        := 2;
   shiftfactor_stereo := 2;
 end;
 meng_proc      := @start_mixen_4;
 nmw_proc       := @nmw_all_4;
 if stereo then
   binnen_proc  := @binnen_loop_4_stereo
 else
   binnen_proc  := @binnen_loop_4;


 seek(rmod,msp+inststart);
 for mlj := 1 to 31 do begin;
  idx := mlj;
  getmem(Samp[idx],Sam_l[idx]);
  blockread(rmod,Samp[idx]^,sam_l[idx]);
 end;

 datgr := inststart - 1083;
 pat_aantal := datgr div Pattgrootte;
 for mlj := 0 to pat_aantal-1 do begin;
   getmem(rm[mlj],2048);
   fillchar(rm[mlj]^,2048,0);
   seek(rmod,msp+1084+mlj*Pattgrootte);
   hulpptr := ptr(seg(rm[mlj]^),ofs(rm[mlj]^));
   for mli := 0 to 63 do begin;
     hulpptr := ptr(seg(rm[mlj]^),ofs(rm[mlj]^)+mli*32);
     blockread(rmod,hulpptr^,Pattgrootte div 64);
   end;
 end;
 seek(rmod,msp+952);
 blockread(rmod,Song,128);

 getmem(strptr,25);
 for i := 0 to 30 do begin;
   seek(rmod,msp+20+i*30);
   blockread(rmod,strptr^,22);
   instnamen[i+1] := convertstring(strptr,22);
 end;
 seek(rmod,msp);
 blockread(rmod,strptr^,20);
 songname := convertstring(strptr,20);
 freemem(strptr,25);

 seek(rmod,msp+950); { von 470}
 blockread(rmod,Songlengte,1);
 end else begin;
 for mlj := 1 to 15 do begin;
  seek(rmod,msp+42+(mlj-1)*30);
  blockread(rmod,sgr,2);
  sgr := swap(sgr) * 2;
  if sgr <> 0 then inststart := inststart - sgr;
  Sam_l[mlj] := sgr;
  seek(rmod,msp+45+(mlj-1)*30);
  blockread(rmod,inst_vol[mlj],1);
  blockread(rmod,loop_s[mlj],2);
  blockread(rmod,loop_l[mlj],2);
  loop_s[mlj] := swap(loop_s[mlj])*2;
  loop_l[mlj] := swap(loop_l[mlj])*2;
 end;

 for mlj := 16 to 31 do begin;
  Sam_l[mlj] := 0;
  loop_s[mlj] := 0;
  loop_l[mlj] := 0;
 end;

 if pos('8CHN',Mkg) <> 0 then begin;
   Pattgrootte    := 2048;
   Stemmen        := 8;
   shiftfactor        := 3;
   shiftfactor_stereo := 3;
 end else begin;
 {4-stemmig MOD-bestand }
   Pattgrootte    := 1024;
   Stemmen        := 4;
   shiftfactor        := 2;
   shiftfactor_stereo := 2;
 end;
 meng_proc      := @start_mixen_4;
 nmw_proc       := @nmw_all_4;
 if stereo then
   binnen_proc     := @binnen_loop_4_stereo
 else
   binnen_proc     := @binnen_loop_4;

 seek(rmod,msp+inststart);
 for mlj := 1 to 15 do begin;
   getmem(Samp[mlj],Sam_l[mlj]);
   blockread(rmod,Samp[mlj]^,sam_l[mlj]);
 end;

 datgr := inststart - 603;
 pat_aantal := datgr div Pattgrootte;
 for mlj := 0 to pat_aantal-1 do begin;
   getmem(rm[mlj],2048);
   fillchar(rm[mlj]^,2048,0);
   seek(rmod,msp+1084+mlj*Pattgrootte);
   hulpptr := ptr(seg(rm[mlj]^),ofs(rm[mlj]^));
   for mli := 0 to 63 do begin;
     hulpptr := ptr(seg(rm[mlj]^),ofs(rm[mlj]^)+mli*32);
     blockread(rmod,hulpptr^,Pattgrootte div 64);
   end;
 end;
 seek(rmod,msp+472);
 blockread(rmod,Song,128);

 getmem(strptr,25);
 for i := 0 to 14 do begin;
   seek(rmod,msp+20+i*30);
   blockread(rmod,strptr^,22);
   instnamen[i+1] := convertstring(strptr,22);
 end;

 for i := 15 to 30 do begin;
   instnamen[i+1] := '';
 end;
 seek(rmod,msp);
 blockread(rmod,strptr^,20);
 songname := convertstring(strptr,20);
 freemem(strptr,25);

 seek(rmod,msp+470);
 blockread(rmod,Songlengte,1);
 end;


 mlj := 0;
 mli := 0;
 close(rmod);
 init_song := true;
end;

procedure exit_song;
begin;
 Port[dsp_adr+$C] := $D3;
 halt(0);
end;

procedure Free_Soundmem;
begin;
 if music_played then begin;
 for mlj := 0 to pat_aantal-1 do begin;
   freemem(rm[mlj],2048);
 end;
 end;
end;

procedure init_sbperiod(p : pointer);
begin;
  periodiek_onderbreken := p;
end;

procedure mod_SetLoopflag(loopen : boolean);
begin;
 mloop := loopen;
end;

procedure mod_SetMultiply(msm : word);
begin;
 modmultiply := msm;
end;

procedure mod_SetLoop(msl : word);
begin;
 Sound_slepen := msl;
 loop3 := msl;
end;

procedure mod_SetSpeed(msp : word);
begin;
 speed := msp;
 Speed3 := msp;
end;

procedure mod_autodetect(what : boolean);
begin;
 if what then mautodet := true else mautodet := false;
end;

procedure mod_transpose(transposerwaarde : integer);
begin;
  tpw := transposerwaarde;
end;

procedure init_data;
Var i,j : integer;
begin;
 m_played := false;
 In_St[1] := 0;
 In_St[2] := 0;
 In_St[3] := 0;
 In_St[4] := 0;
 In_St[5] := 0;
 In_St[6] := 0;
 In_St[7] := 0;
 In_St[8] := 0;
 noot1 := 0;
 noot2 := 0;
 noot3 := 0;
 noot4 := 0;
 noot5 := 0;
 noot6 := 0;
 noot7 := 0;
 noot8 := 0;
 Noot_uitslag[1] := 0;
 Noot_uitslag[2] := 0;
 Noot_uitslag[3] := 0;
 Noot_uitslag[4] := 0;
 Noot_uitslag[5] := 0;
 Noot_uitslag[6] := 0;
 Noot_uitslag[7] := 0;
 Noot_uitslag[8] := 0;
 fillchar(inl,sizeof(inl),0);
 nootvol1 := 0; nootvol2 := 0; nootvol3 := 0; nootvol4 := 0;
 nootvol5 := 0; nootvol6 := 0; nootvol7 := 0; nootvol8 := 0;
 fillchar(Rm_Song,2048,0);
end;

procedure init_Paramtable;
var ls : byte;
    h : real;
begin;
{  playspeed := 6;
  for ls := 1 to 31 do begin;
    if ls <= 3 then
      ModPara[ls].mult := 100
    else
      ModPara[ls].mult := 105;
    ModPara[ls].Speed := ls*ModPara[ls].mult div 10;
    ModPara[ls].bgr := ModPara[ls].Speed*Sound_slepen;
   end;}
end;

procedure mod_Samplefreq(Rate : integer);
var h : real;
begin;
  case Rate of
     08 : begin;
            Sampling_Rate := 131;
            set_timeconst_sb16(131);
            mod_transpose(1);
            mod_SetLoop(15);
            blockgrootte := Speed * Sound_slepen;
            Sampling_Frequentie := 8000;
            init_Paramtable;
          end;
     10 : begin;
            Sampling_Rate := 156;
            set_timeconst_sb16(156);
            mod_transpose(5);
            mod_SetLoop(19);
            blockgrootte := Speed * Sound_slepen;
            Sampling_Frequentie := 10000;
            init_Paramtable;
          end;
     13 : begin;
            Sampling_Rate := 181;
            set_timeconst_sb16(181);
            mod_transpose(10);
            mod_SetLoop(25);
            blockgrootte := Speed * Sound_slepen;
            Sampling_Frequentie := 13333;
            init_Paramtable;
          end;
     16 : begin;
            Sampling_Rate := 196;
            set_timeconst_sb16(196);
            mod_transpose(14);
            mod_SetLoop(32);
            blockgrootte := Speed * Sound_slepen;
            Sampling_Frequentie := 16666;
            init_Paramtable;
          end;
     22 : begin;
            Sampling_Rate := 211;
            set_timeconst_sb16(211);
            mod_transpose(19);
            mod_SetLoop(44);
            blockgrootte := Speed * Sound_slepen;
            Sampling_Frequentie := 22222;
            init_Paramtable;
          end;
   end;
end;

procedure Sound_handler;
var li : integer;
begin;
 if mycli <> 0 then exit;
 mycli := 1;
 if (Loop_pos > Speed) then begin;
   if phase_2 then begin;
     Nothin_done_count := 0;
     asm
       call [nmw_proc]
     end;
     Initialiseer_mengen;
     Loop_pos := 0;
     phase_2 := false;
     phase_1 := true;
     if outfading then
       if outvolume >= 2 then dec(outvolume,2);
     for li := 1 to 8 do
       if Noot_uitslag[li] > 50 then dec(Noot_uitslag[li],50);
   end;
 end else begin;
     asm call [binnen_proc] end;
   Loop_pos := Speed+2;
 end;
 mycli := 0;
end;

procedure calculate_music; assembler;
asm
 cmp mycli,0
 jne  @einde_stop
 cmp  music_uit,0
 jne   @einde_stop
 pusha
 call Sound_handler
 popa
 @einde_stop:
end;

procedure mod_waitretrace(num : byte);
var dl : integer;
begin;
 in_retrace := true;
 for dl := 1 to num do
    calculate_music;
asm
  push dx
@l1:
  mov dx,3dah
  in al,dx
  and al,8h
  jnz @l1
@l2:
  mov dx,3dah
  in al,dx
  and al,8h
  jz @l2
  pop dx
End;
 in_retrace := false;
end;


function laad_modbestand(modname : string;ispeed,iloop : integer;freq : byte) : integer;
var df : file;
    sterreg : byte;
    fgr : longint;
begin;
 PLAYING_MOD := true;
 PLAYING_VOC := false;
 outfading := false;
 outvolume := 63;
 Mod_name := modname;
 {$I-}
 assign(df,Mod_name);
 reset(df,1);
 {$I+}
 if Ioresult <> 0 then begin;
   {$I-}
   close(df);
   laad_modbestand := -1;       {bestand niet gevonden!}
   exit;
 end;
 {$I-}
 fgr := filesize(df);
 close(df);
 music_played := true;
 music_uit := false;
 Mod_afgelopen := false;

 if ispeed <> AUTO then Speed3 := ispeed;
 if iloop <> AUTO then Loop3  := iloop;
 if force_mono then stereo := false;
 if force_sb then begin;
   if Sb16detected then stereo := false;
   Sb16detected := false;
 end;

 if Sbprodetected then begin;
   if stereo then begin;
     sterreg := Read_mixer($0e);
     write_mixer($0e,sterreg OR 2);
   end else begin;
     sterreg := Read_mixer($0e);
     write_mixer($0e,sterreg AND $FD);
   end;
 end;
 init_data;
 if init_song then begin;
   phase_1 := false;
   phase_2 := true;
   mycli := 0;

   mod_Samplefreq(freq);
   Playspeed    := 6;
   Speed := Playspeed*105 div 10;
   bpm := 125;
   mod_SetLoop(Sampling_Frequentie div (BPM * 4));
   blockgrootte := Speed * Sound_slepen;
   if blockgrootte < 100 then blockgrootte := 100;
   if blockgrootte > 4000 then blockgrootte := 4000;

   asm call [nmw_proc] end;
   set_timeconst_sb16(Sampling_Rate);
   Initialiseer_mengen;
   Speelsec := 0;
   Speelmin := 0;
   wr_dsp_sb16($D1);
   if sb16detected or sbprodetected then begin;
     filter_Mid;
     Set_Balance(Balance);
     Set_Volume(Mastervolume);
   end;
   Laad_modbestand := 0;
 end else begin;
   Laad_modbestand := -3;       {fout bij laden van de song}
 end;
end;

procedure einde_mod;
var mlj : integer;
begin;
 Free_Soundmem;
 for mlj := 1 to 31 do begin;
   freemem(Samp[mlj],Sam_l[mlj]);
 end;
 mod_terminated := true;
end;


Procedure periodiek_on;
Begin
 outvolume := 64;
 laatste_weergave := false;
{ for Loop_pos := 1 to Speed do begin;}
   asm call [binnen_proc] end;
{ end;}
 dsp_block_sb16(blockgrootte,blockgrootte,buffer1,true,false);
 bsw := not bsw;
 Loop_pos := 0;
 asm
   call [nmw_proc]
 end;
 Initialiseer_mengen;

 init_sbperiod(@periodiek_off);
 music_played := true;
 TimerAan(@NieuweTimer,timer_per_second);
End;

Procedure periodiek_off;
Begin
  laatste_weergave := true;
  TimerUit;
End;

procedure Fade_Musix_out;
begin;
  outfading := true;
end;

procedure MODExitProc;
var mlj : byte;
 begin
 ExitProc := SaveExitProc;
{ if music_played then periodiek_off;}
if Soundblastergekozeninhoofdprogramma then
   begin
   if not mod_terminated and music_played then einde_mod;
   Exit_Sb16;
   end
end;



{

 **************************************************************************
 ***                                                                    ***
 ***                 Routines voor de weergave van VOC-bestanden        ***
 ***                                                                    ***
 **************************************************************************

}

var pause_voc : boolean;

procedure Init_Voc(filename : string);
const VOCident : string = 'Creative Voice File'+#$1A;
var ch : char;
    ident : string;
    ct : byte;
    h : byte;
    error : integer;
    srlo,srhi : byte;
    SR : word;
    Samplingr : word;
    stereoreg : byte;
begin;
  PLAYING_MOD := false;
  PLAYING_VOC := true;
  VOC_READY   := false;
  vocsstereo := stereo;
  stereo := false;

  assign(vocf,filename);
  reset(vocf,1);
  if filesize(vocf) < 5000 then begin;
    VOC_READY   := true;
    exit;
  end;
  blockread(vocf,voch,$19);
  ident := voch.identstr;
  if ident <> VOCident then begin;
    VOC_READY   := true;
    exit;
  end;

  Blockread(vocf,inread,20);
  vblock.ident := inread[2];

  if vblock.Ident = 1 then begin;
    vblock.SR := inread[6];
  end;

  if vblock.Ident = 8 then begin;
    SR := inread[6]+(inread[7]*256);
    Samplingr := 256000000 div (65536 - SR);
    if inread[9] = 1 then begin; {stereo}
      if sb16detected then samplingr := samplingr shr 1;
      stereo := true;
    end;
    vblock.SR := 256 - longint(1000000 DIV samplingr);
  end;

  if vblock.Ident = 9 then begin;
    Samplingr := inread[6]+(inread[7]*256);
    if inread[11] = 2 then begin; {stereo}
      stereo := true;
      if sbprodetected then samplingr := samplingr * 2;
      vblock.SR := 256 - longint(1000000 DIV (samplingr));
    end else begin;
      vblock.SR := 256 - longint(1000000 DIV samplingr);
    end;
  end;


  if vblock.SR < 130 then vblock.SR := 166;

  set_timeconst_sb16(vblock.SR);

  blockgr := filesize(vocf) - 31;
  if blockgr > 2500 then blockgr := 2500;
  blockread(vocf,vocb1^,blockgr);

  ch := #0;
  fgr := filesize(vocf) - 32;
  fgr := fgr - blockgr;
  Block_activ := 1;

  if fgr > 1 then begin;
    blockread(vocf,vocb2^,blockgr);
    fgr := fgr - blockgr;
  end;

  wr_dsp_sb16($D1);
 lastone := false;

 if not sb16detected then begin;
   if Stereo then begin;
     stereoreg := Read_mixer($0E);
     stereoreg := stereoreg OR 2;
     Write_mixer($0E,stereoreg);
   end else begin;
     stereoreg := Read_mixer($0E);
     stereoreg := stereoreg AND $FD;
     Write_mixer($0E,stereoreg);
   end;
 end;
 pause_voc := false;
 dsp_block_sb16(blockgr,blockgr,vocb1,false,true);
end;

procedure voc_done;
var h : byte;
begin;
 lastone := true;
{ repeat until dsp_rdy_sb16;}
 close(vocf);
 Reset_Sb16;
 stereo := vocsstereo;
end;


procedure voc_pause;
begin;
  pause_voc := true;
end;

procedure voc_continue;
begin;
  pause_voc := false;
  if block_activ = 1 then begin
    dsp_block_sb16(blockgr,blockgr,vocb2,false,true);
    block_activ := 2;
  end else begin;
    dsp_block_sb16(blockgr,blockgr,vocb1,false,true);
    block_activ := 1;
  end;
end;

{

 **************************************************************************
 ***                                                                    ***
 ***    Nog een keer SB - routines, om logische redenen opgenomen       ***
 ***                                                                    ***
 **************************************************************************

}

procedure dsp_int_sb16; interrupt;
var h : byte;
begin;
  if interrupt_check then begin;
    IRQDetected := true;
  end else begin;
    if PLAYING_MOD then begin;
      h := port[dsp_adr+$E];
      dsp_rdy_sb16 := true;

      if not laatste_weergave then begin;
        if bsw then
          dsp_block_sb16(blockgrootte,blockgrootte,buffer1,true,false)
        else
          dsp_block_sb16(blockgrootte,blockgrootte,buffer2,true,false);
        bsw := not bsw;
        phase_1 := false;
        phase_2 := true;
      end;
    end;

    IF PLAYING_VOC then begin;
      h := port[dsp_adr+$E];
      if (fgr > blockgr) and not lastone then begin
        lastone := false;
        if block_activ = 1 then begin
          if not pause_voc then
            dsp_block_sb16(blockgr,blockgr,vocb2,false,true);
          blockread(vocf,vocb1^,blockgr);
          fgr := fgr - blockgr;
          block_activ := 2;
        end else begin;
          if not pause_voc then
            dsp_block_sb16(blockgr,blockgr,vocb1,false,true);
          blockread(vocf,vocb2^,blockgr);
          fgr := fgr - blockgr;
          block_activ := 1;
        end;
      end else begin;
        if not lastone then begin;
          if block_activ = 1 then begin
            if not pause_voc then
              dsp_block_sb16(blockgr,blockgr,vocb2,false,true);
            lastone := true;
          end else begin;
            if not pause_voc then
              dsp_block_sb16(blockgr,blockgr,vocb1,false,true);
            lastone := true;
          end;
        end else begin;
          dsp_rdy_sb16 := true;
          wr_dsp_sb16($D0);
          VOC_READY := true;
        end;
      end;
    end;
  end;
  Port[$20] := $20;
end;

procedure detect_sbIRQ;
const mogelijke_irqs : array[1..5] of byte = ($2,$3,$5,$7,$10);
var i : integer;
    h : byte;
begin;
 getintvec($8+dsp_irq,intback);         { waarden opslaan !          }
 port21 := port[$21];
 fillchar(buffer1^,1200,128);
 set_Timeconst_sb16(211);
 wr_dsp_sb16($D3);                      { luidspreker uit            }
 i := 1;
 interrupt_check := true;
 while (i <= 5) and (not IRQDetected) do
   begin;
     dsp_irq := mogelijke_irqs[i];      { te testen IRQ              }
     getintvec($8+dsp_irq,oldint);      { interrupt omleggen         }
     setintvec($8+dsp_irq,@Dsp_Int_sb16);
     irqmsk := 1 shl dsp_irq;
     port[$21] := port[$21] and not irqmsk;
     Sampling_Rate := 211;
     blockgrootte := 1200;              { testuitvoer                }
     dsp_block_sb16(blockgrootte,blockgrootte,buffer1,true,false);
     delay(150);
     setintvec($8+dsp_irq,oldint);      { interrupt weer terug       }
     port[$21] := Port[$21] or irqmsk;
     h := port[dsp_adr+$E];
     Port[$20] := $20;
     inc(i);
   end;
 interrupt_check := false;
 wr_dsp_sb16($D1);                      { luidspreker weer aan       }
 setintvec($8+dsp_irq,intback);         { oude waarde terug !!!      }
 port[$21] := port21;
 dsp_rdy_sb16 := true;
end;

function Init_Sb : boolean;

begin;
  if not detect_Reg_sb16 then begin;
    Init_Sb := false;
    exit;
  end;
{ Soundblaster gevonden      }
 if not force_irq then detect_sbIRQ;    { IRQ auto-detection         }
   test_overdracht;
 if not force_irq then detect_sbIRQ;    { 2. test voor SB nodig !    }
 if Detect_Mixer_sb16 then begin;
   SbProDetected := TRUE;               { SB Pro gevonden            }
 end;
 SbGetDspVersion;
 if SbVersMaj >= 4 then begin;          { SB 16 ASP gevonden         }
   Sb16Detected := true;
   SBProDetected := false;
 end;
 wr_dsp_sb16($D1);                      { luidspreker aan            }
 getintvec($8+dsp_irq,oldint);          { oude interrupt bewaren,    }
 setintvec($8+dsp_irq,@dsp_int_sb16);   { op eigen routine plaatsen  }
 irqmsk := 1 shl dsp_irq;               { interrupt inmaskeren       }
 port[$21] := port[$21] and not irqmsk;
end;

function init_The_Mod : boolean;
begin;
 mod_autodetect(on);  { indien true is de speed-opgave in de song herkend   }
 mod_SetSpeed(66);
 mod_SetMultiply(11);
 mod_Setloopflag(ON); { moet de weergave weer van voren af aan beginnen,    }
                      { indien afgelopen ?  (ON / OFF)                      }
 if not init_sb then  { speaker inschakelen, automatische herkenning        }
   init_the_mod := false
 else begin;
   init_the_mod := true;
   mod_Samplefreq(Samfreq);
 end;
end;


begin;
 Soundblastergekozeninhoofdprogramma:=false;
 SaveExitProc := ExitProc;
 ExitProc := @MODExitProc;
 dsp_rdy_sb16 := true;
 mod_terminated := false;
 music_played := false;
 mloop := true;
 mli := 0;
 mlj := 0;
 tpw := 5;
 In_St[1] := 0;
 In_St[2] := 0;
 In_St[3] := 0;
 In_St[4] := 0;
 In_St[5] := 0;
 In_St[6] := 0;
 In_St[7] := 0;
 In_St[8] := 0;
 loop_pos := 0;
 mautodet := true;
 modmultiply := 20;
 Sound_slepen := 10;
 Noot_uitslag[1] := 0;
 Noot_uitslag[2] := 0;
 Noot_uitslag[3] := 0;
 Noot_uitslag[4] := 0;
 Noot_uitslag[5] := 0;
 Noot_uitslag[6] := 0;
 Noot_uitslag[7] := 0;
 Noot_uitslag[8] := 0;
 Stemleeg := @lege_stem;
 Mixingprocs[1] := Stemleeg;
 Mixingprocs[2] := Stemleeg;
 Mixingprocs[3] := Stemleeg;
 Mixingprocs[4] := Stemleeg;
 Mixingprocs[5] := Stemleeg;
 Mixingprocs[6] := Stemleeg;
 Mixingprocs[7] := Stemleeg;
 Mixingprocs[8] := Stemleeg;
 getmem(mixed_data,8000);
 getmem(mixed_data_st,8000);

 getmem(buffer1,8000);
 getmem(buffer2,8000);
 getmem(vocb1,8000);
 getmem(vocb2,8000);
 bsw := true;
end.

